# Copyright (c) 2023 ING Analytics Wholesale Banking
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
import psutil
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix

from sparse_dot_topn.lib import _sparse_dot_topn_core as _core
from sparse_dot_topn.types import assert_idx_dtype, assert_supported_dtype, ensure_compatible_dtype

if TYPE_CHECKING:
    from numpy.types import DTypeLike

__all__ = ["sp_matmul", "sp_matmul_topn", "awesome_cossim_topn"]


_N_CORES = psutil.cpu_count(logical=False) - 1

_SUPPORTED_DTYPES = {np.dtype("int32"), np.dtype("int64"), np.dtype("float32"), np.dtype("float64")}


def awesome_cossim_topn(
    A, B, ntop, lower_bound=0, use_threads=False, n_jobs=1, return_best_ntop=None, test_nnz_max=None
):
    """This function will be removed and replaced with `sp_matmul_topn`.

    NOTE this function calls `sp_matmul_topn` but the results may not be the same.
    See the migration guide at 'https://github.com/ing-bank/sparse_dot_topn#migration' for details.

    This function will return a matrix C in CSR format, where
    C = [sorted top n results > lower_bound for each row of A * B].
    If return_best_ntop=True then best_ntop
    (the true maximum number of elements > lower_bound per row of A * B)
    will also be returned in a tuple together with C as (C, best_ntop).

    Args:
        A: LHS of the multiplication, the number of columns of A determines the orientation of B.
            `A` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `B`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        B: RHS of the multiplication, the number of rows of B must match the number of columns of A or the shape of B.T should be match A.
            `B` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `A`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        ntop: top n results
        lower_bound: a threshold that the element of A*B must be greater than
        use_threads: use multi-thread or not
        n_jobs: number of thread, must be >= 1
        return_best_ntop: (default: False) if True, will return best_ntop together
                          with C as a tuple: (C, best_ntop)
        test_nnz_max: deprecated argument, cannot be used

    Returns:
        C: result matrix (returned alone, if return_best_ntop=False)
        best_ntop: The true maximum number of elements > lower_bound per row of
                   A * B returned together with C as a tuple: (C, best_ntop). It is
                   returned only if return_best_ntop=True.

    N.B. if A and B are not in CSR format, they will be converted to CSR
    """
    msg = (
        "`awesome_cossim_topn` function will be removed and (partially) replaced with `sp_matmul_topn`."
        " See the migration guide at 'https://github.com/ing-bank/sparse_dot_topn#readme'."
    )
    if test_nnz_max is not None:
        raise DeprecationWarning(msg)
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    n_threads = n_jobs if use_threads is True else None
    C = sp_matmul_topn(A=A, B=B, top_n=ntop, threshold=lower_bound, sort=True, n_threads=n_threads)
    if return_best_ntop:
        return C, np.diff(C.indptr).max()
    return C


def sp_matmul(
    A: csr_matrix | csc_matrix | coo_matrix,
    B: csr_matrix | csc_matrix | coo_matrix,
    n_threads: int | None = None,
    idx_dtype: DTypeLike | None = None,
) -> csr_matrix:
    """Compute A * B whilst only storing the `top_n` elements.

    This functions allows large matrices to multiplied with a limited memory footprint.

    Args:
        A: LHS of the multiplication, the number of columns of A determines the orientation of B.
            `A` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `B`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        B: RHS of the multiplication, the number of rows of B must match the number of columns of A or the shape of B.T should be match A.
            `B` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `A`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        n_threads: number of threads to use, `None` implies sequential processing, -1 will use all but one of the available cores.
        idx_dtype: dtype to use for the indices, defaults to 32bit integers

    Throws:
        TypeError: when A, B are not trivially convertable to a `CSR matrix`

    Returns:
        C: result matrix

    """
    idx_dtype = assert_idx_dtype(idx_dtype)
    n_threads: int = n_threads or 1
    if n_threads < 0:
        n_threads = _N_CORES

    if isinstance(A, csc_matrix) and isinstance(B, csc_matrix) and A.shape[0] == B.shape[1]:
        A = A.transpose()
        B = B.transpose()
    elif isinstance(A, (coo_matrix, csc_matrix)):
        A = A.tocsr(False)
    elif not isinstance(A, csr_matrix):
        msg = f"type of `A` must be one of `csr_matrix`, `csc_matrix` or `csr_matrix`, got `{type(A)}`"
        raise TypeError(msg)

    if not isinstance(B, (csr_matrix, coo_matrix, csc_matrix)):
        msg = f"type of `B` must be one of `csr_matrix`, `csc_matrix` or `csr_matrix`, got `{type(B)}`"
        raise TypeError(msg)

    A_nrows, A_ncols = A.shape
    B_nrows, B_ncols = B.shape

    if A_ncols == B_nrows:
        if isinstance(B, (coo_matrix, csc_matrix)):
            B = B.tocsr(False)
    elif A_ncols == B_ncols:
        B = B.transpose() if isinstance(B, csc_matrix) else B.transpose().tocsr(False)
        B_nrows, B_ncols = B.shape
    else:
        msg = (
            "Matrices `A` and `B` have incompatible shapes. `A.shape[1]` must be equal to `B.shape[0]` or `B.shape[1]`."
        )
        raise ValueError(msg)

    assert_supported_dtype(A)
    assert_supported_dtype(B)
    ensure_compatible_dtype(A, B)

    # basic check. if A or B are all zeros matrix, return all zero matrix directly
    if A.indices.size == 0 or B.indices.size == 0:
        C_indptr = np.zeros(A_nrows + 1, dtype=idx_dtype)
        C_indices = np.zeros(1, dtype=idx_dtype)
        C_data = np.zeros(1, dtype=A.dtype)
        return csr_matrix((C_data, C_indices, C_indptr), shape=(A_nrows, B_ncols))

    kwargs = {
        "nrows": A_nrows,
        "ncols": B_ncols,
        "A_data": A.data,
        "A_indptr": A.indptr if idx_dtype is None else A.indptr.astype(idx_dtype),
        "A_indices": A.indices if idx_dtype is None else A.indices.astype(idx_dtype),
        "B_data": B.data,
        "B_indptr": B.indptr if idx_dtype is None else B.indptr.astype(idx_dtype),
        "B_indices": B.indices if idx_dtype is None else B.indices.astype(idx_dtype),
    }

    func = _core.sp_matmul
    if n_threads > 1:
        if _core._has_openmp_support:
            kwargs["n_threads"] = n_threads
            func = _core.sp_matmul_mt
        else:
            msg = "sparse_dot_topn: extension was compiled without parallelisation (OpenMP) support, ignoring ``n_threads``"
            warnings.warn(msg, stacklevel=1)
    return csr_matrix(func(**kwargs), shape=(A_nrows, B_ncols))


def sp_matmul_topn(
    A: csr_matrix | csc_matrix | coo_matrix,
    B: csr_matrix | csc_matrix | coo_matrix,
    top_n: int,
    threshold: int | float | None = None,
    sort: bool = False,
    density: float | None = None,
    n_threads: int | None = None,
    idx_dtype: DTypeLike | None = None,
) -> csr_matrix:
    """Compute A * B whilst only storing the `top_n` elements.

    This functions allows large matrices to multiplied with a limited memory footprint.

    Args:
        A: LHS of the multiplication, the number of columns of A determines the orientation of B.
            `A` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `B`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        B: RHS of the multiplication, the number of rows of B must match the number of columns of A or the shape of B.T should be match A.
            `B` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `A`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        top_n: the number of results to retain
        sort: return C in a format where the first non-zero element of each row is the largest value
        threshold: only return values greater than the threshold
        density: the expected density of the result considering `top_n`. The expected number of non-zero elements
            in C should <= (`density` * `top_n` * `A.shape[0]`) otherwise the memory has to reallocated.
            This value should only be set if you have a strong expectation as being wrong incurs a realloaction penalty.
        n_threads: number of threads to use, `None` implies sequential processing, -1 will use all but one of the available cores.
        idx_dtype: dtype to use for the indices, defaults to 32bit integers

    Throws:
        TypeError: when A, B are not trivially convertable to a `CSR matrix`

    Returns:
        C: result matrix

    """
    n_threads: int = n_threads or 1
    if n_threads < 0:
        n_threads = _N_CORES
    density: float = density or 1.0
    idx_dtype = assert_idx_dtype(idx_dtype)

    if isinstance(A, csc_matrix) and isinstance(B, csc_matrix) and A.shape[0] == B.shape[1]:
        A = A.transpose()
        B = B.transpose()
    elif isinstance(A, (coo_matrix, csc_matrix)):
        A = A.tocsr(False)
    elif not isinstance(A, csr_matrix):
        msg = f"type of `A` must be one of `csr_matrix`, `csc_matrix` or `csr_matrix`, got `{type(A)}`"
        raise TypeError(msg)

    if not isinstance(B, (csr_matrix, coo_matrix, csc_matrix)):
        msg = f"type of `B` must be one of `csr_matrix`, `csc_matrix` or `csr_matrix`, got `{type(B)}`"
        raise TypeError(msg)

    A_nrows, A_ncols = A.shape
    B_nrows, B_ncols = B.shape

    if A_ncols == B_nrows:
        if isinstance(B, (coo_matrix, csc_matrix)):
            B = B.tocsr(False)
    elif A_ncols == B_ncols:
        B = B.transpose() if isinstance(B, csc_matrix) else B.transpose().tocsr(False)
        B_nrows, B_ncols = B.shape
    else:
        msg = (
            "Matrices `A` and `B` have incompatible shapes. `A.shape[1]` must be equal to `B.shape[0]` or `B.shape[1]`."
        )
        raise ValueError(msg)

    if B_ncols == top_n and (sort is False) and (threshold is None):
        return sp_matmul(A, B, n_threads)

    assert_supported_dtype(A)
    assert_supported_dtype(B)
    ensure_compatible_dtype(A, B)

    # guard against top_n larger than number of cols
    top_n = min(top_n, B_ncols)

    # handle threshold
    if threshold is not None:
        threshold = int(np.rint(threshold)) if np.issubdtype(A.data.dtype, np.integer) else float(threshold)

    # basic check. if A or B are all zeros matrix, return all zero matrix directly
    if A.indices.size == 0 or B.indices.size == 0:
        C_indptr = np.zeros(A_nrows + 1, dtype=idx_dtype)
        C_indices = np.zeros(1, dtype=idx_dtype)
        C_data = np.zeros(1, dtype=A.dtype)
        return csr_matrix((C_data, C_indices, C_indptr), shape=(A_nrows, B_ncols))

    kwargs = {
        "top_n": top_n,
        "nrows": A_nrows,
        "ncols": B_ncols,
        "threshold": threshold,
        "density": density,
        "A_data": A.data,
        "A_indptr": A.indptr if idx_dtype is None else A.indptr.astype(idx_dtype),
        "A_indices": A.indices if idx_dtype is None else A.indices.astype(idx_dtype),
        "B_data": B.data,
        "B_indptr": B.indptr if idx_dtype is None else B.indptr.astype(idx_dtype),
        "B_indices": B.indices if idx_dtype is None else B.indices.astype(idx_dtype),
    }

    func = _core.sp_matmul_topn if not sort else _core.sp_matmul_topn_sorted
    if n_threads > 1:
        if _core._has_openmp_support:
            kwargs["n_threads"] = n_threads
            kwargs.pop("density")
            func = _core.sp_matmul_topn_mt if not sort else _core.sp_matmul_topn_sorted_mt
        else:
            msg = "sparse_dot_topn: extension was compiled without parallelisation (OpenMP) support, ignoring ``n_threads``"
            warnings.warn(msg, stacklevel=1)
    return csr_matrix(func(**kwargs), shape=(A_nrows, B_ncols))


def zip_sp_matmul_topn(top_n: int, C_mats: list[csr_matrix]) -> csr_matrix:
    """Compute zip-matrix C = zip_i C_i = zip_i A * B_i = A * B whilst only storing the `top_n` elements.

    Combine the sub-matrices together and keep only the `top_n` elements per row.

    Pre-calling this function, matrix B has been split row-wise into chunks B_i, and C_i = A * B_i have been calculated.
    This function computes C = zip_i C_i, which is equivalent to A * B when only keeping the `top_n` elements.
    It allows very large matrices to be split and multiplied with a limited memory footprint.

    Args:
        top_n: the number of results to retain; should be smaller or equal to top_n used to obtain C_mats.
        C_mats: a list with each C_i sub-matrix, with format csr_matrix.

    Returns:
        C: zipped result matrix

    Raises:
        TypeError: when not all elements of `C_mats` is a csr_matrix or trivially convertable
        ValueError: when not all elements of `C_mats` has the same number of rows
    """
    _nrows = []
    ncols = []
    data = []
    indptr = []
    indices = []
    for C in C_mats:
        # check correct type of each C
        if isinstance(C, (coo_matrix, csc_matrix)):
            C = C.tocsr(False)
        elif not isinstance(C, csr_matrix):
            msg = f"type of `C` must be one of `csr_matrix`, `csc_matrix` or `csr_matrix`, got `{type(C)}`"
            raise TypeError(msg)

        nrows, c_nc = C.shape
        _nrows.append(nrows)
        ncols.append(c_nc)
        data.append(C.data)
        indptr.append(C.indptr)
        indices.append(C.indices)

    ncols = np.asarray(ncols, int)
    total_cols = ncols.sum()
    if not np.all(np.diff(_nrows) == 0):
        msg = "Each `C` in `C_mats` should have the same number of rows."
        raise ValueError(msg)

    return csr_matrix(
        _core.zip_sp_matmul_topn(
            top_n=top_n, Z_max_nnz=nrows * top_n, nrows=nrows, B_ncols=ncols, data=data, indptr=indptr, indices=indices
        ),
        shape=(nrows, total_cols),
    )



# Copyright (c) 2023 ING Analytics Wholesale Banking
import importlib.metadata
import os

# Setting the following environment variable allows multiple OpenMP
# libraries to be loaded. This is also used without issue by Scikit-learn.
# OpenMP error msg:
# /* OMP: Error #15: Initializing libomp.dylib, but found libomp.dylib already initialized.
#  * OMP: Hint This means that multiple copies of the OpenMP runtime have been linked into the program.
#  * That is dangerous, since it can degrade performance or cause incorrect results.
#  * The best thing to do is to ensure that only a single OpenMP runtime is linked into the process,
#  * e.g. by avoiding static linking of the OpenMP runtime in any library.
#  * As an unsafe, unsupported, undocumented workaround you can set the environment variable KMP_DUPLICATE_LIB_OK=TRUE
#  * to allow the program to continue to execute, but that may cause crashes or silently produce incorrect results.
#  * For more information, please see http://openmp.llvm.org/
#  */
os.environ.setdefault("KMP_DUPLICATE_LIB_OK", "True")

# Workaround issue discovered in intel-openmp 2019.5:
# https://github.com/ContinuumIO/anaconda-issues/issues/11294
os.environ.setdefault("KMP_INIT_AT_FORK", "FALSE")

__version__ = importlib.metadata.version("sparse_dot_topn")
from sparse_dot_topn.api import awesome_cossim_topn, sp_matmul, sp_matmul_topn, zip_sp_matmul_topn
from sparse_dot_topn.lib import _sparse_dot_topn_core as _core
from sparse_dot_topn.lib._sparse_dot_topn_core import _has_openmp_support

__all__ = [
    "awesome_cossim_topn",
    "sp_matmul",
    "sp_matmul_topn",
    "zip_sp_matmul_topn",
    "_core",
    "__version__",
    "_has_openmp_support",
]



# Copyright (c) 2023 ING Analytics Wholesale Banking
from __future__ import annotations

import psutil
from scipy.sparse import csr_matrix

from sparse_dot_topn.lib import _sparse_dot_topn_core as _core

__all__ = ["sp_matmul_topn"]

_N_CORES = psutil.cpu_count(logical=False) - 1


def sp_matmul_topn(
    A: csr_matrix, B: csr_matrix, top_n: int, sort: bool, threshold: int | float, density: float
) -> csr_matrix:
    """Compute A * B whilst only storing the `top_n` elements.

    This functions allows large matrices to multiplied with a limited memory footprint.

    Note that

    Args:
        A: LHS of the multiplication, the number of columns of A determines the orientation of B.
            `A` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `B`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        B: RHS of the multiplication, the number of rows of B must match the number of columns of A or the shape of B.T should be match A.
            `B` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `A`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        top_n: the number of results to retain
        sort: return C in a format where the first non-zero element of each row is the largest value
        threshold: only return values greater than the threshold, by default this 0.0
        density: the expected density of the result considering `top_n`. The expected number of non-zero elements
            in C should <= (`density` * `top_n` * `A.shape[0]`) otherwise the memory has to reallocated.
            This value should only be set if you have a strong expectation as being wrong incurs a realloaction penalty.

    Returns:
        C: result matrix

    """
    nrows = A.shape[0]
    ncols = B.shape[1]
    func = _core.sp_matmul_topn if not sort else _core.sp_matmul_topn_sorted
    return csr_matrix(
        func(top_n, nrows, ncols, threshold, density, A.data, A.indptr, A.indices, B.data, B.indptr, B.indices),
        shape=(nrows, ncols),
    )


def sp_matmul_topn_mt(
    A: csr_matrix, B: csr_matrix, top_n: int, sort: bool, threshold: int | float, n_threads: int
) -> csr_matrix:
    """Compute A * B whilst only storing the `top_n` elements using multiprocessing.

    This functions allows large matrices to multiplied with a limited memory footprint.

    Note that

    Args:
        A: LHS of the multiplication, the number of columns of A determines the orientation of B.
            `A` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `B`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        B: RHS of the multiplication, the number of rows of B must match the number of columns of A or the shape of B.T should be match A.
            `B` must be have an {32, 64}bit {int, float} dtype that is of the same kind as `A`.
            Note the matrix is converted (copied) to CSR format if a CSC or COO matrix.
        top_n: the number of results to retain
        sort: return C in a format where the first non-zero element of each row is the largest value
        threshold: only return values greater than the threshold, by default this 0.0
        n_threads: number of threads to use, -1 will use all but one of the available cores.

    Returns:
        C: result matrix

    """
    nrows = A.shape[0]
    ncols = B.shape[1]
    n_threads = n_threads if n_threads > 0 else _N_CORES
    func = _core.sp_matmul_topn_mt if not sort else _core.sp_matmul_topn_sorted_mt
    return csr_matrix(
        func(top_n, nrows, ncols, threshold, n_threads, A.data, A.indptr, A.indices, B.data, B.indptr, B.indices),
        shape=(nrows, ncols),
    )



import sys
from itertools import product

import numpy as np
import pytest
from scipy import sparse
from sparse_dot_topn import _has_openmp_support, sp_matmul, sp_matmul_topn, zip_sp_matmul_topn

from ._resources import _assert_array_equal, _assert_smat_equal, _get_topn_elements


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_default(rng, dtype):
    A = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    C = sp_matmul(A, B)
    C_ref = A.dot(B)
    _assert_smat_equal(C, C_ref)


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_nthreads(rng, dtype):
    A = sparse.random(100, 10, density=0.5, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(10, 100, density=0.5, format="csr", dtype=dtype, random_state=rng)
    C_ref = A.dot(B)
    if not _has_openmp_support:
        with pytest.warns(UserWarning):
            C = sp_matmul(A, B, n_threads=2)
    else:
        C = sp_matmul(A, B, n_threads=2)
    _assert_smat_equal(C, C_ref)


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_default(rng, dtype):
    A = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    C = sp_matmul_topn(A, B, top_n=B.shape[1])
    C_ref = A.dot(B)
    _assert_smat_equal(C, C_ref)


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_topn(rng, dtype):
    A = sparse.random(100, 10, density=0.5, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(10, 100, density=0.5, format="csr", dtype=dtype, random_state=rng)
    C_ref = A.dot(B)
    C_10 = sp_matmul_topn(A, B, top_n=10)
    C_30 = sp_matmul_topn(A, B, top_n=30)
    for i in range(A.shape[0]):
        _assert_array_equal(C_10[i, :].data, _get_topn_elements(C_ref[i, :].data, 10))
        _assert_array_equal(C_30[i, :].data, _get_topn_elements(C_ref[i, :].data, 30))


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_nthreads(rng, dtype):
    A = sparse.random(100, 10, density=0.5, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(10, 100, density=0.5, format="csr", dtype=dtype, random_state=rng)
    C_ref = A.dot(B)
    if not _has_openmp_support:
        with pytest.warns(UserWarning):
            C = sp_matmul_topn(A, B, top_n=B.shape[1], n_threads=2)
    else:
        C = sp_matmul_topn(A, B, top_n=B.shape[1], n_threads=2)
    _assert_smat_equal(C, C_ref)

    C_10 = sp_matmul_topn(A, B, top_n=10, n_threads=2)
    C_30 = sp_matmul_topn(A, B, top_n=30, n_threads=2)
    for i in range(A.shape[0]):
        _assert_array_equal(C_10[i, :].data, _get_topn_elements(C_ref[i, :].data, 10))
        _assert_array_equal(C_30[i, :].data, _get_topn_elements(C_ref[i, :].data, 30))


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_sorted(rng, dtype):
    A = sparse.random(100, 10, density=0.5, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(10, 100, density=0.5, format="csr", dtype=dtype, random_state=rng)
    C_ref = A.dot(B)

    C_10 = sp_matmul_topn(A, B, top_n=10, sort=True)
    C_30 = sp_matmul_topn(A, B, top_n=30, sort=True, n_threads=2)
    for i in range(A.shape[0]):
        sorted_row = np.sort(C_ref[i, :].data)[::-1]
        _assert_array_equal(C_10[i, :].data, sorted_row[:10])
        _assert_array_equal(C_30[i, :].data, sorted_row[:30])


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_density(rng, dtype):
    A = sparse.random(200, 200, density=0.9, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(200, 200, density=0.9, format="csr", dtype=dtype, random_state=rng)
    C = sp_matmul_topn(A, B, top_n=B.shape[1], density=0.01)
    C_ref = A.dot(B)
    _assert_smat_equal(C, C_ref)


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_threshold(rng, dtype):
    A = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    threshold = 0.0 if np.issubdtype(A.data.dtype, np.integer) else 0.5
    C = sp_matmul_topn(A, B, top_n=B.shape[1], threshold=threshold)
    C_ref = A.dot(B)
    assert C.data.min() > threshold
    assert C.data.size < C_ref.data.size


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_sp_matmul_topn_idx_dtype(rng, dtype):
    A = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(100, 100, density=0.1, format="csr", dtype=dtype, random_state=rng)
    C = sp_matmul_topn(A, B, top_n=B.shape[1], idx_dtype=np.int64)
    C_ref = A.dot(B)
    _assert_smat_equal(C, C_ref)


_FORMATS = ["coo", "csr", "csc"]


@pytest.mark.parametrize("formats", product(_FORMATS, _FORMATS))
def test_sp_matmul_topn_formats(rng, formats):
    A = sparse.random(200, 100, density=0.1, format=formats[0], random_state=rng)
    B = sparse.random(100, 300, density=0.1, format=formats[1], random_state=rng)

    C = sp_matmul_topn(A, B, top_n=B.shape[1])
    C_ref = A.tocsr().dot(B.tocsr())
    _assert_smat_equal(C, C_ref)


def test_sp_matmul_topn_transpose(rng):
    A = sparse.random(200, 100, density=0.1, format="csr", random_state=rng)
    B = sparse.random(100, 300, density=0.1, format="csr", random_state=rng)

    C = sp_matmul_topn(A.transpose(), B.transpose(), top_n=B.shape[1])
    C_ref = A.dot(B)
    _assert_smat_equal(C, C_ref)


_KWARGS = [
    {
        "A": {"m": 10, "n": 10, "density": 0.1, "format": "csr"},
        "B": {"m": 10, "n": 10, "density": 0.1, "format": "csr", "T": False},
    },
    {
        "A": {"m": 100, "n": 100, "density": 0.1, "format": "csr"},
        "B": {"m": 100, "n": 100, "density": 0.1, "format": "csr", "T": False},
    },
    {
        "A": {"m": 1000, "n": 1000, "density": 0.05, "format": "csr"},
        "B": {"m": 1000, "n": 1000, "density": 0.05, "format": "csr", "T": False},
    },
    {
        "A": {"m": 2000, "n": 2000, "density": 0.01, "format": "csr"},
        "B": {"m": 2000, "n": 2000, "density": 0.01, "format": "csr", "T": False},
    },
    {
        "A": {"m": 10000, "n": 10000, "density": 0.001, "format": "csr"},
        "B": {"m": 10000, "n": 10000, "density": 0.001, "format": "csr", "T": False},
    },
    {
        "A": {"m": 1000, "n": 100, "density": 0.1, "format": "csr"},
        "B": {"m": 100, "n": 100, "density": 0.1, "format": "csr", "T": False},
    },
    {
        "A": {"m": 1000, "n": 100, "density": 0.1, "format": "csr"},
        "B": {"m": 100, "n": 20, "density": 0.1, "format": "csr", "T": False},
    },
    {
        "A": {"m": 1000, "n": 100, "density": 0.1, "format": "csr"},
        "B": {"m": 200, "n": 100, "density": 0.1, "format": "csr", "T": True},
    },
    {
        "A": {"m": 1000, "n": 100, "density": 0.1, "format": "csr"},
        "B": {"m": 200, "n": 100, "density": 0.1, "format": "csc", "T": True},
    },
]


@pytest.mark.parametrize("kwargs", _KWARGS)
def test_sp_matmul_topn_broadcasting(rng, kwargs):
    A = sparse.random(**kwargs["A"], random_state=rng)
    T_b = kwargs["B"].pop("T")
    B = sparse.random(**kwargs["B"], random_state=rng)

    topn = B.shape[0] if T_b else B.shape[1]
    C_ref = A.dot(B.T) if T_b else A.dot(B)
    C = sp_matmul_topn(A, B, top_n=topn)
    _assert_smat_equal(C, C_ref)


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_zip_sp_matmul_topn(rng, dtype):
    # matching 100 names against 600 gt-names, where gt has been split into three parts
    A = sparse.random(100, 2000, density=0.1, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(600, 2000, density=0.1, format="csr", dtype=dtype, random_state=rng)

    # reference
    C_ref = sp_matmul_topn(A, B.T, top_n=10, threshold=0.01, sort=True)

    # zipped C-matrix
    Bs = [B[:100], B[100:300], B[300:]]
    Cs = [sp_matmul_topn(A, Bi.T, top_n=10, threshold=0.01, sort=True) for Bi in Bs]
    C_zip = zip_sp_matmul_topn(top_n=10, C_mats=Cs)

    # comparison of index-pointers, data, indices
    _assert_array_equal(C_zip.indptr, C_ref.indptr)
    _assert_array_equal(C_zip.data, C_ref.data)
    # NB indices are not necessarily equal. For example in case of multiple equal scores close to the topn boundary,
    # sp_matmul_topn and the zipped approach may pick different matches. In such cases for sp_matmul_topn
    # the insertion order in the maxheap is leading, which is impossible to replicate in zip_sp_matmul_topn,
    # as the B matrices are have been split and get inserted in separate maxheap objects.
    _assert_array_equal(C_zip.indices, C_ref.indices)


@pytest.mark.skipif(sys.version_info < (3, 9), reason="not all dtypes supported in scipy vstack for python 3.8")
@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
def test_stack_zip_sp_matmul_topn(rng, dtype):
    # matching 1000 names against 600 gt-names,
    # where gt has been split into three parts, and names-to-match in five parts
    A = sparse.random(1000, 2000, density=0.1, format="csr", dtype=dtype, random_state=rng)
    B = sparse.random(600, 2000, density=0.1, format="csr", dtype=dtype, random_state=rng)

    # reference
    C_ref = sp_matmul_topn(A, B.T, top_n=10, threshold=0.01, sort=True)

    # zipped C-matrix
    As = [A[i * 200 : (i + 1) * 200] for i in range(5)]  # names-to-match split
    Bs = [B[:100], B[100:300], B[300:]]  # GT split

    Cs = [[sp_matmul_topn(Aj, Bi.T, top_n=10, threshold=0.01, sort=True) for Bi in Bs] for Aj in As]
    C_zips = [zip_sp_matmul_topn(top_n=10, C_mats=Cis) for Cis in Cs]

    # stack over names-to-match subparts
    C_stack = sparse.vstack(C_zips, dtype=dtype)

    # comparison of index-pointers, data, indices
    _assert_array_equal(C_stack.indptr, C_ref.indptr)
    _assert_array_equal(C_stack.data, C_ref.data)
    # NB indices are not necessarily equal. For example in case of multiple equal scores close to the topn boundary,
    # sp_matmul_topn and the zipped approach may pick different matches. In such cases for sp_matmul_topn
    # the insertion order in the maxheap is leading, which is impossible to replicate in zip_sp_matmul_topn,
    # as the B matrices are have been split and get inserted in separate maxheap objects.
    _assert_array_equal(C_stack.indices, C_ref.indices)



# Copyright (c) 2023 ING Analytics Wholesale Banking
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.types import DTypeLike, NDArray
    from scipy.sparse import coo_matrix, csc_matrix, csr_matrix

__all__ = ["assert_idx_dtype", "assert_supported_dtype", "ensure_compatible_dtype", "is_supported_dtype"]

_SUPPORTED_DTYPES = {np.dtype("int32"), np.dtype("int64"), np.dtype("float32"), np.dtype("float64")}


def assert_idx_dtype(dtype: DTypeLike | None) -> DTypeLike:
    if dtype is None:
        return np.dtype("int32")
    if not (dtype == np.dtype("int32") or dtype == np.dtype("int64")):
        msg = f"`idx_dtype` must be a 32 or 64 bit integer, got {dtype}."
        raise TypeError(msg)
    return dtype


def assert_supported_dtype(obj: NDArray | coo_matrix | csc_matrix | csr_matrix, name: str | None = None):
    if obj.dtype in _SUPPORTED_DTYPES:
        return
    msg = "Supported dtypes are {32, 64}bit {int, float}" + f" got {name or 'obj'}.dtype: {obj.dtype}"
    raise TypeError(msg)


def ensure_compatible_dtype(
    a: NDArray | coo_matrix | csc_matrix | csr_matrix, b: NDArray | coo_matrix | csc_matrix | csr_matrix
):
    """Check and convert dtype if needed and safe to do.

    When the dtypes match nothing is done.
    If one has a lower precision dtype of the same kind as the other, the lower precision object
    is cast to the higher precision one. Otherwhise an error is raised.

    """
    if a.dtype == b.dtype:
        return a, b
    if a.dtype.itemsize > b.dtype.itemsize:
        lhs = b
        rhs = a
    else:
        lhs = a
        rhs = b
    if np.issubdtype(lhs.dtype, rhs.dtype):
        return lhs.astype(rhs.dtype), rhs
    msg = "`a` and `b` do not have the same dtype and cannot be safely cast"
    raise TypeError(msg)


def is_supported_dtype(dtype: DTypeLike) -> bool:
    return dtype in _SUPPORTED_DTYPES



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

if TYPE_CHECKING:
    from numpy.types import NDArray


def _get_topn_elements(x: NDArray, n: int):
    return x[np.sort(np.argsort(x)[::-1][:n])]


def _assert_array_equal(A, B, rtol=1e-5, atol=1e-8):
    if np.issubdtype(A.dtype, np.integer):
        assert_array_equal(A, B)
    else:
        assert_allclose(A, B, rtol=rtol, atol=atol)


def _assert_smat_equal(A, B, rtol=1e-5, atol=1e-8):
    _assert_array_equal(A.data, B.data, rtol, atol)
    assert_array_equal(A.indptr, B.indptr)
    assert_array_equal(A.indices, B.indices)



# Copyright (c) 2023 ING Analytics Wholesale Banking
from __future__ import annotations

import re
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.sparse import load_npz, save_npz
from sklearn.feature_extraction.text import TfidfVectorizer
from sparse_dot_topn import sp_matmul, sp_matmul_topn

# A.shape: 331500, 193190
# B.shape: 331500, 193190

# Max number of rows possible is 331500
N_ROWS = 20_000
INPUT_FILE = "sec__edgar_company_info.csv.gz"

dir_path = Path(__file__).parent.resolve()
path_A = dir_path / f"tfidf_A_{N_ROWS}.npz"
path_B = dir_path / f"tfidf_B_{N_ROWS}.npz"

if not (path_A.exists() and path_B.exists()):

    def ngrams(string, n=3):
        string = re.sub(r"[,-./]|\sBD", r"", string)
        ngrams = zip(*[string[i:] for i in range(n)])
        return ["".join(ngram) for ngram in ngrams]

    dir_path = Path(__file__).parent.resolve()
    names = pd.read_csv(dir_path / INPUT_FILE)
    company_names = names["Company Name"].str.lower().sample(frac=1).reset_index(drop=True)
    del names

    hn = company_names.size // 2

    vectorizer = TfidfVectorizer(min_df=1).fit(company_names)
    A = vectorizer.transform(company_names[:hn])
    B = vectorizer.transform(company_names[hn:])

    rng = np.random.default_rng()

    A_srows = np.sort(rng.choice(np.arange(hn), size=N_ROWS, replace=False))
    B_srows = np.sort(rng.choice(np.arange(hn), size=N_ROWS, replace=False))

    save_npz(path_A, A[A_srows, :])
    save_npz(path_B, B[B_srows, :].transpose().tocsr())

A = load_npz(path_A)
B = load_npz(path_B)


def bench_scipy_csr():
    A.dot(B)


__benchmarks__ = [
    (bench_scipy_csr, partial(sp_matmul, A, B, n_threads=1), "Scipy vs sp_matmul                    | n_threads: 1"),
    (bench_scipy_csr, partial(sp_matmul, A, B, n_threads=2), "Scipy vs sp_matmul                    | n_threads: 2"),
    (bench_scipy_csr, partial(sp_matmul, A, B, n_threads=4), "Scipy vs sp_matmul                    | n_threads: 4"),
    (bench_scipy_csr, partial(sp_matmul, A, B, n_threads=8), "Scipy vs sp_matmul                    | n_threads: 8"),
    (bench_scipy_csr, partial(sp_matmul_topn, A, B, 10), "Scipy vs sp_matmul_topn | top_n: 10   | n_threads: 1"),
    (bench_scipy_csr, partial(sp_matmul_topn, A, B, 20), "Scipy vs sp_matmul_topn | top_n: 20   | n_threads: 1"),
    (bench_scipy_csr, partial(sp_matmul_topn, A, B, 30), "Scipy vs sp_matmul_topn | top_n: 30   | n_threads: 1"),
    (bench_scipy_csr, partial(sp_matmul_topn, A, B, 100), "Scipy vs sp_matmul_topn | top_n: 100  | n_threads: 1"),
    (bench_scipy_csr, partial(sp_matmul_topn, A, B, 1000), "Scipy vs sp_matmul_topn | top_n: 1000 | n_threads: 1"),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 10, n_threads=2),
        "Scipy vs sp_matmul_topn | top_n: 10   | n_threads: 2",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 20, n_threads=2),
        "Scipy vs sp_matmul_topn | top_n: 20   | n_threads: 2",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 30, n_threads=2),
        "Scipy vs sp_matmul_topn | top_n: 30   | n_threads: 2",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 100, n_threads=2),
        "Scipy vs sp_matmul_topn | top_n: 100  | n_threads: 2",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 1000, n_threads=2),
        "Scipy vs sp_matmul_topn | top_n: 1000 | n_threads: 2",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 10, n_threads=4),
        "Scipy vs sp_matmul_topn | top_n: 10   | n_threads: 4",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 20, n_threads=4),
        "Scipy vs sp_matmul_topn | top_n: 20   | n_threads: 4",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 30, n_threads=4),
        "Scipy vs sp_matmul_topn | top_n: 30   | n_threads: 4",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 100, n_threads=4),
        "Scipy vs sp_matmul_topn | top_n: 100  | n_threads: 4",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 1000, n_threads=4),
        "Scipy vs sp_matmul_topn | top_n: 1000 | n_threads: 4",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 10, n_threads=8),
        "Scipy vs sp_matmul_topn | top_n: 10   | n_threads: 8",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 20, n_threads=8),
        "Scipy vs sp_matmul_topn | top_n: 20   | n_threads: 8",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 30, n_threads=8),
        "Scipy vs sp_matmul_topn | top_n: 30   | n_threads: 8",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 100, n_threads=8),
        "Scipy vs sp_matmul_topn | top_n: 100  | n_threads: 8",
    ),
    (
        bench_scipy_csr,
        partial(sp_matmul_topn, A, B, 1000, n_threads=8),
        "Scipy vs sp_matmul_topn | top_n: 1000 | n_threads: 8",
    ),
]
